local OO = MiTu.using(MiTu.OO)
MiTu.CastMonitor = MiTu.CastMonitor or {}
local CurrentPackage = MiTu.BeginPackage(MiTu.CastMonitor)

--------------------------------------------------------------------------------
--############################################################################--
--------------------------------------------------------------------------------
OO.NewClass(CurrentPackage, "CastMonitorFrame", 
    function(self, owner)
    	
    	self.owner = owner
    	self.frame = nil
    	self.items = {}
    	
    	self.itemPool = {}
    	self:initFrame()
    	self:initEvents()
    	
    end
)

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:onResize()
	-- TODO
end

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:onEvent(eventType, ...)
	if (eventType == "RAID_ROSTER_UPDATE") then
		-- TODO: Debugausgabe
		--DEFAULT_CHAT_FRAME:AddMessage("RosterUpdate")
		self:updateItems()
	end
end

--------------------------------------------------------------------------------
-- Looks through all items in the list and removes items associated with
-- characters that are no longer in the raid
-- also updates the unitids of all remaining items
--------------------------------------------------------------------------------
function CastMonitorFrame:updateItems()

	for index, item in ipairs(self.items) do
		local unitName = item:getUnitName()
		local unitID = self.owner:getRaidUnitID(unitName)
		if (not unitID) then
			self:removeItem(item)
		else
			-- TODO: Debugausgabe
			item:setUnit(unitID)
			--DEFAULT_CHAT_FRAME:AddMessage(item:getUnitName().."-"..item:getUnit())
		end
	end
end

--------------------------------------------------------------------------------
-- Makes the frame visible. 
--------------------------------------------------------------------------------
function CastMonitorFrame:show()
	self.frame:Show()
end
--------------------------------------------------------------------------------
-- Hides the frame
--------------------------------------------------------------------------------
function CastMonitorFrame:hide()
	self.frame:Hide()
end
--------------------------------------------------------------------------------
-- Toggles the visibility of the frame
--------------------------------------------------------------------------------
function CastMonitorFrame:toggle()
	if (self.frame:IsShown()) then
		self:hide()
	else
		self:show()
	end
end

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:getFrame()
	return self.frame
end

--------------------------------------------------------------------------------
-- Called whenever the user clicks the "+"-Button to add a character to the 
-- list.
--------------------------------------------------------------------------------
function CastMonitorFrame:onAddItem()
	self.owner:addCurrentTarget()
end

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:initFrame()

	local me = self
	
	self.frame = CreateFrame("FRAME", "CastMonitorMainFrame", UIParent)
	self.frame:SetPoint("CENTER", UIParent, "CENTER")
	self.frame:SetWidth(300)
	self.frame:SetHeight(200)
	self.frame:EnableMouse(true)
    self.frame:SetMovable(true)
    self.frame:SetResizable(true)
    
    self.frame:SetScale(CastMonitorSettings.Frame.Scale)
	self.frame:RegisterForDrag("LeftButton") 
    self.frame:SetScript("OnDragStart", function() this:StartMoving() end)
    self.frame:SetScript("OnDragStop", function() this:StopMovingOrSizing() end)
    self.frame:SetScript("OnSizeChanged", function() me:onResize() end)
	self.frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		tile = true, 
		tileSize = 16,
		edgeFile = "Interface\\Buttons\\WHITE8X8.blp", 
		edgeSize = 1,
		insets = {left = 1, right = 1, top = 1, bottom = 1},
	})
	self.frame:SetBackdropColor(0, 0, 0, 0.75)
	self.frame:SetBackdropBorderColor(0, 0, 0, 1)
	
	self.addButton = CreateFrame("Button", nil, self.frame, 
		"UIPanelButtonTemplate")
	self.addButton:ClearAllPoints()
	self.addButton:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT")
	self.addButton:SetWidth(CastMonitorSettings.Frame.FontSize
		+ CastMonitorItem.Padding)
	self.addButton:SetHeight(CastMonitorSettings.Frame.FontSize
		+ CastMonitorItem.Padding)
	self.addButton:Show()
	self.addButton:SetText("+")
	self.addButton:SetScript("OnClick", function()
		me:onAddItem()
	end)
	
	self:initDragHandles()
	
	self.frame:Hide()
	
end

--------------------------------------------------------------------------------
-- Creates and initializes the handles that are used to resize the frame.
-- Called by "initFrame" once when setting up the frame.
--------------------------------------------------------------------------------
function CastMonitorFrame:initDragHandles()

	self.frame.DragBottomRight = CreateFrame("Frame", nil, self.frame)
	self.frame.DragBottomRight:Show()
	self.frame.DragBottomRight:SetFrameLevel( self.frame:GetFrameLevel() + 10)
	self.frame.DragBottomRight:SetWidth(16)
	self.frame.DragBottomRight:SetHeight(16)
	self.frame.DragBottomRight:SetPoint("BOTTOMRIGHT", 
		self.frame, "BOTTOMRIGHT", 0, 0)
	self.frame.DragBottomRight:EnableMouse(true)
	self.frame.DragBottomRight:SetScript("OnMouseDown", 
		function() 
			this:GetParent().isResizing = true; 
			this:GetParent():StartSizing("BOTTOMRIGHT") 
		end 
	)
	self.frame.DragBottomRight:SetScript("OnMouseUp", 
		function() 
			this:GetParent():StopMovingOrSizing(); 
			this:GetParent().isResizing = false; 
		end 
	)
	
end
--------------------------------------------------------------------------------
function CastMonitorFrame:removeItem(toBeRemoved)
	for index, item in ipairs(self.items) do
		if item == toBeRemoved then
			tremove(self.items, index)
			item:hide()
			tinsert(self.itemPool, item)
			self:doLayout()
			return
		end
	end
end

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:addItem(unitID)

	-- Before adding the item, cleanup the current list
	self:updateItems()
	
	-- chech wether the item is already in the list	
	if (self:isUnitInList(unitID)) then 
		DEFAULT_CHAT_FRAME:AddMessage(
			"CastMonitor: Canno add unit. It is already in the list")
		return
	end
	
	local color = nil
	if (UnitName(unitID) == UnitName("player")) then
		color = CastMonitorSettings.CastbarsColors.Player
	else
		color = CastMonitorSettings.CastbarsColors.Other
	end
	
	local item = nil
	if (#self.itemPool > 0) then
		item = tremove(self.itemPool)
		item:setUnit(unitID)
		item:setCastbarColor(color)
		item:show()
	else 
		item = CastMonitorItem(self, unitID, color)
	end
	tinsert(self.items, item)
		
	self:doLayout()
	
end
--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:isUnitInList(unitID)
	for _, item in pairs(self.items) do
		if (item:getUnit() == unitID) then
			return true
		end
	end
	return false
end
--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:doLayout()

	local height = 0
	local isFirst = true
	local currentFrame = nil
	local previousFrame = nil
	
	
	for _, item in pairs(self.items) do
		currentFrame = item:getFrame()
		currentFrame:ClearAllPoints()
		height = height + currentFrame:GetHeight()
		if (isFirst == true) then
			isFirst = false
			currentFrame:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
			currentFrame:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT")
		else
			currentFrame:SetPoint("TOPLEFT", previousFrame, "BOTTOMLEFT")
			currentFrame:SetPoint("TOPRIGHT", previousFrame, "BOTTOMRIGHT")
		end
		previousFrame = currentFrame
	end
	
	-- TODO: Die 16 hier nicht als Konstante. Ist die Hhe der Draghandles 
	-- am unteren Rand
	self.frame:SetHeight(height + 16)
	self.frame:SetMinResize(CastMonitorSettings.Frame.SourceFrameWidth
		+ CastMonitorSettings.Frame.TargetFrameWidth + 16, height + 16)
	
end
--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
function CastMonitorFrame:initEvents()
	
	local me = self
	
	self.frame:SetScript("OnEvent", function(self, event,...)
		me:onEvent(event,...)
	end)
	
	self.frame:RegisterEvent("RAID_ROSTER_UPDATE")
end
--------------------------------------------------------------------------------