local OO = MiTu.using(MiTu.OO)
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
MiTu.CastMonitor = MiTu.CastMonitor or {}
local CurrentPackage = MiTu.BeginPackage(MiTu.CastMonitor)

--------------------------------------------------------------------------------
--############################################################################--
--------------------------------------------------------------------------------
OO.NewClass(CurrentPackage, "CastMonitorItem", 
    function(self, castMonitorFrame, unit, castbarColor)
    
    	self.castMonitorFrame = castMonitorFrame
    	
    	self.castProgress = 0
    	self.castStartTime = 0
    	self.castTime = 0
    	self.castEndTime = 0
    	self.spellName = ""
    	self.spellRank = ""
    	self.spellLabel = ""
    	self.spellIconPath = ""
    	self.isCasting = false
    	self.fadeElapsed = CastMonitorSettings.Behavior.FadeTime
    	self.castbarColor = castbarColor
    	self.unitName = ""
    	
    	self:initFrame()
    	self:setUnit(unit)
    	self:initEvents()
    	
    	self:update()
    	
    end
)
--------------------------------------------------------------------------------
--Static Stuff
--------------------------------------------------------------------------------
CastMonitorItem.Padding = 4

--------------------------------------------------------------------------------
-- Called, whenever the unit associated with this item starts casting a spell
--------------------------------------------------------------------------------
function CastMonitorItem:onSpellcastStart()

	self:initCast()
	self.isCasting = true
	
end
--------------------------------------------------------------------------------
-- Called, whenever the unit associated with this item stops casting a spell
-- this is:
-- cancelling a cast (player and raid members)
-- finishing a cast (player)
--------------------------------------------------------------------------------
function CastMonitorItem:onSpellcastStop(unit, spellName, spellRank)
	
	self.isCasting = false	
	-- TODO: self or not self, that is the question
	self:startFading(unit, self.spellName, self.spellRank)
end

--------------------------------------------------------------------------------
-- starts fading out the castbar.
--------------------------------------------------------------------------------
function CastMonitorItem:startFading(unit, spellName, spellRank)
	
	local castText = spellName 
	
	if (spellRank ~= nil) then
		castText = castText .."  ("..spellRank..")"
	end
	-- TODO: Move constant to config
	self.castFrame:SetAlpha(1)
	self:setCastbarText(castText)
	self.fadeElapsed = 0
end
--------------------------------------------------------------------------------
-- initializes or updates the castbar.
-- This is called whenever the unit assciated with this item starts casting a 
-- spell or if the spell is delayed and the castbar needs to be updated
--------------------------------------------------------------------------------
function CastMonitorItem:initCast()

	self.spellName, self.spellRank, self.spellLabel, self.spellIconPath, 
		self.castStartTime, self.castEndTime = UnitCastingInfo(self.unit);
	
	self.castTime = self.castEndTime - self.castStartTime;
	local castText = self.spellName
	if (self.spellRank ~= nil) then
		castText = castText .."  ("..self.spellRank..")"
	end
	self.fadeElapsed = 1
	self:setCastbar(0, self.castTime, 0, castText)
	self.castbar:SetStatusBarColor(
		self.castbarColor.r, self.castbarColor.g, self.castbarColor.b, 1)
end
--------------------------------------------------------------------------------
-- Sets minValue, maxValue, value and text for the castbar
--------------------------------------------------------------------------------
function CastMonitorItem:setCastbar(minValue, maxValue, value, text)

	self.castbar:SetMinMaxValues(minValue, maxValue)
	self.castbar:SetValue(value)
	self.castFrame:SetAlpha(1)
	self:setCastbarText(text)
end
--------------------------------------------------------------------------------
-- sets the text displayed in the castbar
--------------------------------------------------------------------------------
function CastMonitorItem:setCastbarText(text)
	self.castString:SetText(text)
end

--------------------------------------------------------------------------------
-- Called, whenever the cast of the unit associated with this item is delayed
-- updates the castbar
--------------------------------------------------------------------------------
function CastMonitorItem:onSpellcastDelayed()

	if (not self.isCasting) then
		-- Should'nt happen, but who knows?
		return
	end 
	
	-- recalculate the cast
	self:initCast()
		
end
--------------------------------------------------------------------------------
-- Called, whenever the cast of the unit associated with this item is finished
-- (cast time finished or spell failed)
-- starts fading out the castbar
--------------------------------------------------------------------------------
function CastMonitorItem:onSpellcastFinished(unit, spellName, spellRank)
	self.isCasting = false	
	self:startFading(unit, spellName, spellRank)
end
--------------------------------------------------------------------------------
-- Called, whenever a cast succeeds (either an instant cast or a normal cast)
-- Starts fading the castbar
--------------------------------------------------------------------------------
function CastMonitorItem:onSpellcastSucceeded(unit, spellName, spellRank)

	self.isCasting = false
	self.castbar:SetMinMaxValues(0, 100)
	self.castbar:SetValue(100)
	self.castbar:SetStatusBarColor(
		CastMonitorSettings.CastbarsColors.Finished.r, 
		CastMonitorSettings.CastbarsColors.Finished.g,
		CastMonitorSettings.CastbarsColors.Finished.b
	)
	self:startFading(unit, spellName, spellRank)
end

--------------------------------------------------------------------------------
-- Called, whenever the target of the unit changes.
-- Updates the target display
--------------------------------------------------------------------------------
function CastMonitorItem:onTargetChanged()
	self:update()
end

--------------------------------------------------------------------------------
-- Updates the target display (name, reaction color, class color)
--------------------------------------------------------------------------------
function CastMonitorItem:update()
	--update the target
	if (self.targetFrame ~= nil) then
	
		if UnitExists(self.unit.."target") then
		
			-- target foreground
			local targetName = UnitName(self.unit.."target")
			self.targetFrame.Text:SetText(targetName)
			local _, className = UnitClass(self.unit.."target")
			local color = RAID_CLASS_COLORS[className]
			if (not color) then
				color = {r = 1, g = 1, b = 1}
			end
			self.targetFrame.Text:SetTextColor(color.r, color.g, color.b)
			
			--target background
			color = CastMonitor.getReactionColor(
				self.unit, self.unit.."target") 
			self.targetFrame:SetBackdropColor(
				color.r, color.g, color.b, color.a
			)
			
		else 
			self.targetFrame.Text:SetText("(none)")
			self.targetFrame.Text:SetTextColor(1, 1, 1)
			
			local color = CastMonitorSettings.TargetColors.None
			self.targetFrame:SetBackdropColor(
				color.r, color.g, color.b, color.a
			)
		end
	end
end
--------------------------------------------------------------------------------
-- called automatically for all registered events
-- Dispatches the event to corresponding method
--------------------------------------------------------------------------------
function CastMonitorItem:onEvent(eventType, ...)
	
	local unit = select(1, ...)
		
	-- if this event is fore some other unit than the one we're 
	-- interested:ignore
	-- except we have  a COMBAT_LOG_EVENT_UNFILTERED event, then the unit is
	-- checked in "onCombatLogEvent"
	if (eventType ~= "COMBAT_LOG_EVENT_UNFILTERED" 
		and (unit ~= self.unit)) then
		return
	end
	
	if (eventType == "UNIT_SPELLCAST_START") then
		self:onSpellcastStart(...)
	elseif (eventType == "UNIT_SPELLCAST_STOP"
		or eventType == "UNIT_SPELLCAST_INTERRUPTED") then
		-- called for player and raid, when spell is stopped  
		-- not called for raid member, if spell finishes
		self:onSpellcastStop(...)
	elseif eventType == "UNIT_SPELLCAST_DELAYED" then
		self:onSpellcastDelayed()
	elseif eventType == "UNIT_SPELLCAST_SUCCEEDED" then
		--This is handled by COMBAT_LOG_EVENT_UNFILTERED:SPELL_CAST_SUCCESS
		--self:onSpellcastSucceeded(...)
	elseif eventType == "UNIT_SPELLCAST_FAILED" then
		self:onSpellcastFinished(...)
	elseif (eventType == "UNIT_TARGET") then
		self:onTargetChanged(...)
	elseif (eventType == "COMBAT_LOG_EVENT_UNFILTERED") then
		self:onCombatLogEvent(...)
	end
end
--------------------------------------------------------------------------------
-- handles combat log events.
-- this is used to properly detect instant casts from raid members
--------------------------------------------------------------------------------
function CastMonitorItem:onCombatLogEvent(...)

	local timestamp, event, srcGUID, srcName, srcFlags, dstGUID, 
		dstName, dstFlags, value, unknown1 = select(1, ...);
	
	
	if (srcName ~= self.unitName) then
		return
	end
	
	if (event ~= "SPELL_CAST_SUCCESS") then
		return
	end
	
	local _, spellName, _ = select(9, ...);
	
	self:onSpellcastSucceeded(nil, spellName, nil)
	
end
--------------------------------------------------------------------------------
-- called ever frame
-- updates castbar lengths and castbar fade value 
--------------------------------------------------------------------------------
function CastMonitorItem:onUpdate(elapsed)

	-- if self.fadeElapsed is lower than the fade time fading has to be done
	-- to the bar
	if (self.fadeElapsed < CastMonitorSettings.Behavior.FadeTime) then
		self.fadeElapsed = (self.fadeElapsed + elapsed);
		self.castFrame:SetAlpha(1 
			- (self.fadeElapsed / CastMonitorSettings.Behavior.FadeTime))
	end

	-- no need to update castbars if not casting
	if (not self.isCasting) then
		return
	end

	self.castProgress = (GetTime() * 1000) - self.castStartTime
	if (self.castProgress >= self.castTime) then
		self.castProgress = self.castTime;
		self:onSpellcastFinished(self.unit, self.spellName, self.spellRank)
		return
	end
	
	self.castbar:SetValue(self.castProgress)
end

--------------------------------------------------------------------------------
-- Returns the frame object associated with this item.
-- Usually used by the mainwindow to position all items correctly
--------------------------------------------------------------------------------
function CastMonitorItem:getFrame()
	return self.frame
end
--------------------------------------------------------------------------------
function CastMonitorItem:hide()
	self.frame:Hide()
end
--------------------------------------------------------------------------------
function CastMonitorItem:show()
	self.frame:Show()
end
--------------------------------------------------------------------------------
function CastMonitorItem:getUnit()
	return self.unit
end
--------------------------------------------------------------------------------
function CastMonitorItem:setUnit(unitID)
	self.unit = unitID
	if (self.unit ~= nil) then
		local name, realm = UnitName(self.unit)
    	self.unitName = name
    	if (realm and (realm ~= "")) then
    		self.unitName = self.unitName.."-"..realm
    	end
    else
    	self.unitName = ""
    end
    if (self.sourceFrame ~= nil) then
    	local _, className = UnitClass(self.unit)
		local color = RAID_CLASS_COLORS[className]
			self.sourceFrame.Text:SetTextColor(color.r, color.g, color.b)
    	self.sourceFrame.Text:SetText(self.unitName)
    end
    
    -- the unit changed so there is probably a different target as well
    -- update it
    self:update()
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:setCastbarColor(color)
	self.castbarColor = color
	
end
--------------------------------------------------------------------------------
-- Called whenever the user clicks this items "remove button".
-- Tells the mainwindow to remove this item
--------------------------------------------------------------------------------
function CastMonitorItem:onRemoveItem()

	-- not unregistering because frame is probably reused
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_START")
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_STOP")
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_DELAYED")
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	--self.frame:UnregisterEvent("UNIT_SPELLCAST_FAILED")
	--self.frame:UnregisterEvent("UNIT_TARGET")
		
	self.castMonitorFrame:removeItem(self)
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:initFrame()
	
	self.frame = CreateFrame("FRAME", nil, self.castMonitorFrame:getFrame())
	self.frame:SetHeight(CastMonitorSettings.Frame.FontSize
		+ CastMonitorItem.Padding)
		
	self:initRemoveButton()
	self:initSourceFrame()
	self:initTargetFrame()
	self:initCastFrame()
	
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:initRemoveButton()

	local me = self
	
	self.removeButton = CreateFrame("Button", nil, self.frame, 
		"UIPanelButtonTemplate")
	self.removeButton:ClearAllPoints()
	self.removeButton:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
	self.removeButton:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT")
	self.removeButton:SetWidth(self.frame:GetHeight())
	self.removeButton:Show()
	self.removeButton:SetText("-")
	
	self.removeButton:SetScript("OnClick", function()
		me:onRemoveItem()
	end)
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:initSourceFrame()
	self.sourceFrame = CreateFrame("Frame", nil, self.frame)
	self.sourceFrame:SetWidth(CastMonitorSettings.Frame.SourceFrameWidth)
	self.sourceFrame:SetPoint("TOPLEFT", self.removeButton, "TOPRIGHT")
	self.sourceFrame:SetPoint("BOTTOMLEFT", self.removeButton, "BOTTOMRIGHT")
	--
	--self.sourceFrame:SetBackdrop({
	--	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
	--	tile = true, 
	--	tileSize = 16
	--})
	--self.frame:SetBackdropColor(0, 0, 0, 0.75)
	
	self.sourceFrame:EnableMouse(true)
	self.sourceFrame:SetScript("OnEnter", function()
		--me:showTooltip(this)
	end)
	self.sourceFrame:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
			
	self.sourceFrame.Text = self.sourceFrame:CreateFontString(
		nil, "OVERLAY", "GameFontNormal")
	self.sourceFrame.Text:SetAllPoints(self.sourceFrame)
	self.sourceFrame.Text:SetJustifyH("LEFT")
	self.sourceFrame.Text:SetTextColor(1.0, 1.0, 1.0)
	self.sourceFrame.Text:SetFont("Fonts\\FRIZQT__.TTF", 
		CastMonitorSettings.Frame.FontSize)
	self.sourceFrame.Text:SetNonSpaceWrap(false)
	-- TODO: Text Testweise
	self.sourceFrame.Text:SetText(self.unitName)
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:initCastFrame()
	self.castFrame = CreateFrame("Frame", nil, self.frame)
	self.castFrame:SetPoint("TOPLEFT", self.sourceFrame, "TOPRIGHT")
	self.castFrame:SetPoint("BOTTOMRIGHT", self.targetFrame, "BOTTOMLEFT")
	
	self.castbar = CreateFrame("StatusBar", nil, self.castFrame)
	self.castbar:ClearAllPoints()
	self.castbar:SetPoint("TOPLEFT", self.castFrame, "TOPLEFT")
	self.castbar:SetPoint("BOTTOMRIGHT", self.castFrame, "BOTTOMRIGHT")
	--self.castbar:SetHeight(10)
	
	self.castbar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
	self.castbar:SetStatusBarColor(
		self.castbarColor.r, self.castbarColor.g, self.castbarColor.b, 1)
	self.castbar:SetMinMaxValues(0, 100)
	self.castbar:SetValue(0)
	
	self.castString = self.castbar:CreateFontString(nil, 
		"OVERLAY", "GameFontNormal")
	self.castString:SetAllPoints(self.castFrame)
	self.castString:SetJustifyH("LEFT")
	self.castString:SetTextColor(1.0, 1.0, 1.0)
	self.castString:SetFont("Fonts\\FRIZQT__.TTF", 
		CastMonitorSettings.Frame.FontSize)
	self.castString:SetNonSpaceWrap(false)
	-- TODO: Text Testweise
	self.castString:SetText("")
	
end
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
function CastMonitorItem:initTargetFrame()
	self.targetFrame = CreateFrame("Frame", nil, self.frame)
	self.targetFrame:SetHeight(CastMonitorSettings.Frame.FontSize)
	self.targetFrame:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT")
	self.targetFrame:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT")
	self.targetFrame:SetWidth(CastMonitorSettings.Frame.TargetFrameWidth)
	self.targetFrame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		tile = true, 
		tileSize = 16,
		--edgeFile = "Interface\\Buttons\\WHITE8X8.blp", 
		--edgeSize = 1,
		--insets = {left = 1, right = 1, top = 1, bottom = 1},
	})
	local color = CastMonitorSettings.TargetColors.None
	self.targetFrame:SetBackdropColor(
		color.r, color.g, color.b, color.a
	)
	
	self.targetFrame:EnableMouse(true)
	self.targetFrame:SetScript("OnEnter", function()
		--me:showTooltip(this)
	end)
	self.targetFrame:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
			
	self.targetFrame.Text = self.sourceFrame:CreateFontString(
		nil, "OVERLAY", "GameFontNormal")
	self.targetFrame.Text:SetAllPoints(self.targetFrame)
	self.targetFrame.Text:SetJustifyH("LEFT")
	self.targetFrame.Text:SetTextColor(1.0, 1.0, 1.0)
	self.targetFrame.Text:SetFont("Fonts\\FRIZQT__.TTF", 
		CastMonitorSettings.Frame.FontSize)
	self.targetFrame.Text:SetNonSpaceWrap(false)
	-- TODO: Text Testweise
	self.targetFrame.Text:SetText("(none)")
end
--------------------------------------------------------------------------------
-- Initializes all events needed for this item.
-- Only called once when this item is created
--------------------------------------------------------------------------------
function CastMonitorItem:initEvents()
	local me = self
	self.frame:SetScript("OnEvent", function(self, event,...)
		me:onEvent(event,...)
	end)
	self.frame:SetScript("OnUpdate", function(self, elapsed)
		me:onUpdate(elapsed)
	end)
	
	self.frame:RegisterEvent("UNIT_SPELLCAST_START")
	self.frame:RegisterEvent("UNIT_SPELLCAST_STOP")
	self.frame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self.frame:RegisterEvent("UNIT_SPELLCAST_DELAYED")
	--self.frame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	--self.frame:RegisterEvent("UNIT_SPELLCAST_FAILED")
	self.frame:RegisterEvent("UNIT_TARGET")
	self.frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	
	
end
--------------------------------------------------------------------------------
-- Returns this items unit name
--------------------------------------------------------------------------------
function CastMonitorItem:getUnitName()
	return self.unitName
end