
--############################################################################--
--------------------------------------------------------------------------------
--                                                                            --
--                                                                            --
--   #####            #####    ####    ################    ####        ####   --
--   ########      ########    ####    ################    ####        ####   --
--   ####  ####  ####  ####    ####          ####          ####        ####   --
--   ####     ####     ####    ####          ####          ####        ####   --
--   ####      ##      ####    ####          ####          ####        ####   --
--   ####              ####    ####          ####           #####    #####    --
--   ####              ####    ####          ####             ##########      --
--------------------------------------------------------------------------------
--   WoW-Addon Framework by Merdon of Milites Turris@Die Silberne Hand (EU)   --
--------------------------------------------------------------------------------
--############################################################################--
--##                                                                        ##--
--## QUICKSTART                                                             ##--
--## ----------                                                             ##--
--############################################################################--
--##                                                                        ##--
--## 1. Declaring Classes:                                                  ##--
--## ---------------------                                                  ##--
--## NewClass(Package, "ClassName", BaseClass, ConstructorFunc)             ##--
--##                                                                        ##--
--## Examples:                                                              ##--
--## a) simple class, no inheritance, default constructor:                  ##--
--## MyPackage = {}                                                         ##--
--## NewClass(MyPackage, "Object")                                          ##--
--## Your Package MyPackage will now contain a (Class-)Table called Object  ##--
--## The function NewClass also returns the Class so you could write        ##--
--## Object = NewClass(Mypackage, "Object")                                 ##--
--## use this, if you need the new class in a diffrent environment than     ##--
--## your current                                                           ##--
--##                                                                        ##--
--## b) inheritance, default constructor                                    ##--
--## NewClass(MyPackage, "Control", Object)                                 ##--
--## The current environment will contain a (Class-)Table called Control    ##--
--## With Object as its base class                                          ##--
--## Again, you can (but do not need to) write                              ##--
--## Control = NewClass("Control", Object)                                  ##--
--##                                                                        ##--
--## c) with constructor                                                    ##--
--## NewClass(MyPackage, "Button", Control, function(self, text) {          ##--
--##     Button.base.constructor(self)                                      ##--
--##     self.text = text                                                   ##--
--## })                                                                     ##--
--## the constructor function must include the "self" parameter at the      ##--
--## beginning of the parameter list. You must also call the base class'    ##--
--## constructor manually (only) if your class has a base class             ##--
--## (using "." notation and passing self as first parameter)               ##--
--## If you do not specify a constructor a default constructor              ##--
--## will be created. The default constructor will automatically call       ##--
--## the base class' constructor (if there is a base class)                 ##--
--##                                                                        ##--
--## 2. Declaring Methods                                                   ##--
--## --------------------                                                   ##--
--## function ClassName:Methodname(parameter list)                          ##--
--##     --method body                                                      ##--
--## end                                                                    ##--
--##                                                                        ##--
--## Examples:                                                              ##--
--## a) simple method, simple parameter                                     ##--
--## funcion Botton:SetText(text)                                           ##--
--##     self.text = text                                                   ##--
--## end                                                                    ##--
--## Note that "self" automatically refers to the class instance because    ##--
--## we used ":" notation  when declaring our method                        ##--
--##                                                                        ##--
--## b) calling base class' method                                          ##--
--## function Button:SetWidth(width)                                        ##--
--##     Button.base.SetWidth(self, width)                                  ##--
--##     -- some extra work...                                              ##--
--## end                                                                    ##--
--## Note, that you have to use "." notation and pass self as first         ##--
--## parameter if you are you are calling base class methods                ##--
--##                                                                        ##--
--## 3. Creating instances                                                  ##--
--## ---------------------                                                  ##--
--## var = ClassName(parameterlist)                                         ##--
--##                                                                        ##--
--## Examples:                                                              ##--
--## local myButton = Button("Cancel")                                      ##--
--## All parameters will be passed to your constructor                      ##--
--## or the default constructor                                             ##--
--############################################################################--

MiTu = MiTu or {}   -- namespace for alle MiTu-Packages

-- Starting own package / environment
MiTu.OO = {}
local CurrentPackage = MiTu.BeginPackage(MiTu.OO)

--############################################################################--
--##                                                                        ##--
--############################################################################--
local function validateClassName(className)
    
    local _G = getfenv()
    
    if (type(className) ~= "string") then
        -- TODO: Fehler: Klassenname muss ein String sein
    end

    if (_G[className]) then
        -- TODO. Fehler: Klassenname existiert bereits im globalen Environment
    end

    -- TODO: berprfen, ob der String ungltige Zeichen enthlt
    -- dem angegebenen Namen besitzt
end

--############################################################################--
--##                                                                        ##--
--############################################################################--
function NewClass(package, className, base, constructor)
    
    -- TODO: Fehler ausgeben, wenn Klassenname ungltig
    --validateClassName(className)

    local class = {} -- Tabelle, die die Klasse reprsentiert. Sie wird am Ende der Methode 'NewClass' zurckgegeben
    local mt = {} -- Metatabelle der Klasse. 
    setmetatable(class, mt)
	class.metatable = mt

    class.name = className -- Die Klasse soll ihren Namen kennen. Deshalb wird er im Feld "name" gespeichert
    
    -- Wenn constructor keine gltige function aber base, dann hat die Klasse keine Basisklasse und der Konstruktor
    -- wurde stattdessen im Parameter base bergeben
    if (((not constructor) or (type(constructor) ~= "function" )) and type(base) == "function") then
        constructor = base
        base = nil
    end 


    if (base ~= nil) then
        if (type(base) ~= "table") then
            -- TODO: Fehler: Basisklasse muss eine Tabelle sein
        end

        class.base = base --jede Klasse soll eine direkte Referenz auf die basisklasse haben
        mt.__index = base --Felder und Methoden, die in der Klasse nicht gefunden werden sollen in der Basisklasse gesucht werden. Deshalb wird __index in der Metatabelle auf die Basisklasse gesetzt
    end
    class.__index = class --Die Klasse ist Metatabelle fr ihre Instanzen, aber gleichzeitig auch der Index. Deshalb wird hier __index (class als Metatabelle) auf class selbst gesetzt (class als Index)

    -- Einen Standardkontstruktor erzeugen, falls keiner bergeben wurde
    if (constructor == nil) then
        local defaultConstructor = function(instance)
            --Falls eine Basisklasse existiert, dann den Konstruktor der Basisklasse aufrufen.
            --instance ist die Tabelle des zu "konstruierenden" Objektes
            if (base ~= nil) then
                base.constructor(instance)
            end
        end
        class.constructor = defaultConstructor
    else
        class.constructor = constructor -- Die Klasse merkt sich den Konstruktor fr den spteren Gebrauch
    end

    -- Neue Instanzen der Klasse sollen im Stil von "var = Klassenname()" erzeugt werden knnen. Deshalb in der Metatabelle der Klasse  "__call" auf eine
    -- Funktion setzen, die das Erzeugen der neuen Instanz bernimmt
    mt.__call = function(classTable,...) 
        local arg = arg or {} -- Falls dem Konstruktor keine Parameter bergeben wurden, dann erstellen wir eine leere Parameterliste
        local instance = {} -- Diese Tabelle reprsentiert die neue Instanz
        instance.class = class -- Damit die Klasse einer Tabelle leicht ermittelt werden kann wird sie im Feld "class" der Instanz gespeichert
        instance.base = base -- Um das Aufrufen von Methoden der Basisklasse zu erleichtern wird die basisklasse auch direkt in der Instanz gespeichert
        setmetatable(instance, class) -- Methoden und Felder, die nicht in der Instanz gefunden werden, sollen in der Klasse gesucht werden. Deshalb wird hier die Klasse als Metatabelle gesetzt (zur Erinnerung: die Klassentabelle enthlt ein Feld "__index", das wiederum auf die Klasse selbst zeigt
        
        class.constructor(instance, ...) -- Den Konstruktor aufrufen, um "instance" zu initialieiseren
        return instance
    end 
    -- Eine Variable mit dem Klassennamen als Variablennamen im Paket erzeugen, 
    -- falls ein Paket bergeben wurde, damit die Klassentabelle direkt 
    -- verwendet werden kann, anstatt der Klassenname als String
    if (package ~= nil) then
        package[className] = class
    else
        local _G = getfenv()
        _G[className] = class 
    end
    
    return class
end

-- TODO: Die __tostring-Methode muss noch berarbeitet werden.   




